/*

  Linux driver for BCI-2003, BCI-2004, BCI-2103, BCI-2104 Q-bus/Unibus
  interface boards (made by The Logical Company).

  By John Wilson <wilson@dbit.com>.

  Copyright (C) 2000-2006 by Digby's Bitpile, Inc.  All rights reserved.

  This driver is licensed under GPLv2, and may be distributed only in
  accordance with its terms.  The above copyright notice must be preserved in
  all copies and derivative works (and displayed in the Linux system log), and
  source code must be available.

  03/11/2000	JMBW	Created.
  03/12/2000	JMBW	I/O page accesses and interrupts work, mostly.
  03/13/2000	JMBW	DMA works (BCI-2004).
  08/24/2001	JMBW	Fixed kernel-version-specific conditionals, added
			support for Unibus map.  V1.1.
  03/12/2002	JMBW	Updated build procedure for 2.4.5 kernel.
  04/24/2004	JMBW	Added BCIDELAY to fix UMR problem.  V1.2.
  02/23/2006	JMBW	Updated to work with 2.6.x kernel + hotplug + udev.
			BCIBRAB ioctl added, BCIREQS fixed.  V1.3.

  Bus adapters made by The Logical Company:

  BCI-2000	PCI/Unibus, not emulator-friendly, NOT SUPPORTED
  BCI-2003	PCI/Unibus, contiguous DMA
  BCI-2004	PCI/Unibus, mapped DMA
  BCI-2100	PCI/Q-bus, not emulator-friendly, NOT SUPPORTED
  BCI-2103	PCI/Q-bus, contiguous DMA
  BCI-2104	PCI/Q-bus, mapped DMA

  This file must be compiled with -O in order to remove the unused cases in
  get_user/put_user.  Otherwise you get an undefined external error for
  __get_user_X since there's no such routine.

  OK the new style of Makefile takes care of that for us.  It depends on
  /lib/module/x.y.z/build containing (or linking to) the kernel sources.
  "make modules_install" takes care of this.
  The udev stuff comes from:
	http://www.kernel.org/pub/linux/utils/kernel/hotplug

*/

#define NAME "bci"		/* name of device driver */
#define BCIMAJOR 218		/* HPA assigned major device # */
#define BCIVERSION "1.3"	/* version # of this driver */

#define PCI_VENDOR_ID_TLC 0x1356  /* The Logical Company's PCI vendor ID */
#define BCI2x03 0x000C		/* BCI-2003, BCI-2103 product ID */
#define BCI2x04 0x0013		/* BCI-2004, BCI-2104 product ID */

#define SIZE2x03 0x00000200UL	/* size of BCI2003/2103 register window */
#define SIZE2x04 0x00020000UL	/* size of BCI2004/2104 register window */

/* offsets in BAR0/BAR1 (ba->port) I/O-mapped space */
#define PLXDMC 0x28		/* direct master control reg */
#define PLXICSR 0x68		/* interrupt control/status reg */
#define PLXREG 0x70		/* size of local config I/O port region */

/* offsets in BAR2 (ba->regs) memory-mapped space */
#define BCICSR 0x00		/* control/status reg */
#define BCIBDR 0x08		/* bus data reg */
#define BCIBVR 0x0C		/* bus vector reg */
#define BCIBPR 0x10		/* bus priority reg */
#define BCIBMR 0x14		/* Q/Unibus maintenance reg */
#define BCIBAR 0x18		/* bus addr reg */
#define BCIOFR 0x1C		/* base addr offset reg (BCI-2x03) */
#define BCISGC 0x1C		/* scatter/gather CSR (BCI-2x04) */
#define BCIMCR 0x20		/* mapping control reg (BCI-200x only) */
#define BCITPR 0x24		/* test point register */

#define BCIUMR 0x100		/* Unibus map registers (32. longwords) */

#define BCISGM 0x10000		/* scatter/gather map regs (8192. quadwords) */

#define BA ((struct barec *)ba)	/* cast "ba" correctly when it's "void *" */

#define BCIDELAY ndelay(500)	/* delay after touching registers */
				/* (some writes, like UMRs, won't stick if */
				/* you write too soon after previous) */

#include <linux/delay.h>	/* udelay(), ndelay() */
#include <linux/device.h>	/* class stuff */
#include <linux/err.h>		/* IS_ERR(), PTR_ERR() macros */
#include <linux/errno.h>	/* "Exxxx" constants */
#include <linux/fs.h>		/* file_operations, kill_fasync() etc. */
#include <linux/init.h>		/* module_init(), module_exit() */
#include <linux/ioport.h>	/* request_(mem_)region(), release_XXX() */
#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* module-related macros */
#include <linux/page-flags.h>	/* PageReserved(), SetPageDirty() macros */
#include <linux/pagemap.h>	/* page_cache_entry() macro */
#include <linux/pci.h>		/* pci_dev, pci*() */
#include <linux/sched.h>	/* request_irq(), free_irq() */
#include <linux/slab.h>		/* kmalloc(), kfree() */
#include <linux/string.h>	/* memcpy(), memset(), strncmp() */
#include <linux/version.h>	/* LINUX_VERSION_CODE, KERNEL_VERSION */
#include <linux/cdev.h>		/* cdev stuff (nds on above files, which?) */
#include <asm/scatterlist.h>	/* struct scatterlist */
#include <asm/io.h>		/* ioremap_nocache(), iounmap(), ioread*(), */
				/* iowrite*() */
#include <asm/uaccess.h>	/* get_user(), put_user() */

MODULE_AUTHOR("John Wilson <wilson@dbit.com>");
MODULE_DESCRIPTION("D Bit driver for Logical BCI-2xxx bus adapters");
MODULE_LICENSE("GPL");

#include "bci.h"		/* ioctl() stuff */

/* bus adapter information record, one per BCI board */
struct barec {
	struct barec *next;	/* pointer to next in BALIST, or NULL */
	unsigned short devid;	/* PCI device ID (BCI2x03 or BCI2x04) */
	unsigned int unit;	/* which BCI this is (0 = first) */
	unsigned char inuse;	/* NZ => in use, i.e. open()ed by someone */
	unsigned char qbus;	/* NZ => Q-bus, 0 => Unibus */
	unsigned char mapped;	/* NZ => scatter/gather map, 0 => contig */
	unsigned char irqreg;	/* NZ => we've registered our IRQ */
	unsigned char memreg;	/* NZ => we've registered our memory ports */
	unsigned char portreg;	/* NZ => we've registered our I/O ports */
	unsigned char cdevreg;	/* NZ => we've registered our struct cdev */
	unsigned char classreg;	/* NZ => we've registered our class */
	unsigned char soft;	/* NZ => software only (no BCI hardware) */
				/* (for debugging) */
	unsigned int port;	/* PLX PCI9060 I/O port addresses */
	unsigned long regphys, reglen;  /* phys addr/len of BAR2 reg block */
	unsigned int irq;	/* IRQ # */
	unsigned int csr;	/* BCICSR value */
	void *regs;		/* ioremap_nocache() window to BAR2 BCI regs */
				/* and map registers */
	struct fasync_struct *fasync_queue;
				/* head of FASYNC notification queue */
	volatile unsigned char br4[4];
				/* flags for each bus request prio level, */
				/* each var contains its own bit # if that */
				/* BR needs probing */
#define BR4 br4[0]		/* 0x10 => BR4 needs probing */
#define BR5 br4[1]		/* 0x20 => BR5 needs probing */
#define BR6 br4[2]		/* 0x40 => BR6 needs probing */
#define BR7 br4[3]		/* 0x80 => BR7 needs probing */
				/* (this way if you OR them all together, */
				/* you get a bitmap) */
#define DMASIZE 4194304		/* size of DMA area (4 MB) */
#define MAXPAGE (DMASIZE/PAGE_SIZE)  /* # pages in DMA area */
	unsigned int offset;	/* offset (in bytes) of DMA base from start */
				/* of DMA page buffer (E11 always leaves */
				/* offset=0 starting with V4.0) */
	unsigned long pages[MAXPAGE];
				/* DMA pages -- first OFFSET bytes not used */
				/* (but available to client as regular RAM) */
	struct page *pgs[MAXPAGE];
				/* new with V1.3:  page structures for user */
				/* memory (ioctl(BCIDMA)) */
	unsigned int npgs;	/* # pages currently mapped */
	struct scatterlist sg[MAXPAGE];
				/* scatter/gather descriptor list */
	unsigned int nsg;	/* # scatter/gather descriptors */
	struct pci_dev *pcidev;	/* ptr to our pci_dev struct (unless SOFT) */
	struct cdev cdev;	/* our char dev struct */
/// not yet used
};

static struct barec *balist = NULL;  /* list of all BCI boards, in order */
// do we need a spinlock protecting this during link/unlink/search?

/* list of PCI devices that are our business: */
static __devinitdata struct pci_device_id bcitable[] = {
	{ PCI_DEVICE(PCI_VENDOR_ID_TLC,BCI2x03) },
	{ PCI_DEVICE(PCI_VENDOR_ID_TLC,BCI2x04) },
	{ 0, }
};
MODULE_DEVICE_TABLE(pci,bcitable);

/* insmod parameter to create software bus adapter (for debugging) */
static int soft=0;
module_param(soft,int,0444);
MODULE_PARM_DESC(soft,"1 => create one software (fake) bus adapter");

/* class structure for bus adapters */
/* (we need this to show up in /sys/class so that udev will see us) */
/* I hate the ever-changing Linux device driver interface SO MUCH */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,2)
/* no class stuff before 2.6.2 (not the same kind, anyway) */
# define CLASSDEVCREATE(a,b,c,d) do {} while(0)
# define CLASSDEVDESTROY(a,b) do {} while(0)
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,12)
/* 2.6.2 - 2.6.12 */
static struct class_simple *bciclass;
# define CLASSDEVCREATE(a,b,c,d) class_simple_device_add(a,b,NULL,c,d)
# define CLASSDEVDESTROY(a,b) class_simple_device_remove(b)
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,14)
/* 2.6.13, 2.6.14 */
static struct class *bciclass;
# define CLASSDEVCREATE(a,b,c,d) class_device_create(a,b,NULL,c,d)
# define CLASSDEVDESTROY(a,b) class_device_destroy(a,b)
#else
/* 2.6.15 (how long is this likely to stay working?!) */
static struct class *bciclass;
# define CLASSDEVCREATE(a,b,c,d) class_device_create(a,NULL,b,NULL,c,d)
# define CLASSDEVDESTROY(a,b) class_device_destroy(a,b)
#endif

/* I have fantasies of writing Linux drivers for Q/Unibus devices, which */
/* will call this driver to access them */
static int bcimatch(struct device *dev,struct device_driver *driver)
{
	return(0);
}

static int bcihotplug(struct device *dev,char **envp,int nenvp,char *buf,
	int bufsize)
{
//
	return(0);
}

/* vectors for bus operations, called by kernel */
static struct bus_type bcibus = {
	.name = "unibus",	/* (or Q-bus, but I had to pick one name) */
	.match = bcimatch,
	.hotplug = bcihotplug,
};

#if 0 /* some day */
int unibus_register_device(struct ubdev *ubdev)
{
}
EXPORT_SYMBOL(unibus_register_device);

void unibus_unregister_device(struct ubdev *ubdev)
{
}
EXPORT_SYMBOL(unibus_unregister_device);

int unibus_register_driver(struct ubdrv *ubdrv)
{
}
EXPORT_SYMBOL(unibus_register_driver)

void unibus_unregister_driver(struct ubdrv *ubdrv)
{
}
EXPORT_SYMBOL(unibus_unregister_driver);
#endif /* 0 */

/* interrupt service routine */
static irqreturn_t bciisr(int irq,void *ba,struct pt_regs *regs)
{
	register unsigned int n;
	/* check LI bit to make sure the interrupt is ours (if shared IRQ) */
	if(!(inl(BA->port+PLXICSR)&0x00008000)) return(IRQ_NONE);
	BCIDELAY;

	/* what we'd really like to do here is have individual control over */
	/* the four levels, so that we'd get an int on any leading edge */

	/* but the hardware is trying to act like a real PDP-11 and we can */
	/* disable only from a given priority on down, we can't disable only */
	/* the BRs that we already know about and get new ints on the others */

	/* we can't just pass the PDP-11's priority through to the hardware */
	/* because it will be out of sync with the PDP-11 if an int comes in */
	/* part way through the emulation of an instruction which would have */
	/* resulted in that priority level being masked */

	/* the solution is, when a BR is asserted, we flag that BR and all */
	/* lower levels as requiring service, and then increase the BCI's */
	/* priority register so that BR (and unfortunately the lower ones) */
	/* won't cause the ISR to loop (due to PCI ints being level- */
	/* triggered) */

	/* once Ersatz-11 is informed of the BR level(s) that need service */
	/* (one for sure, better safe than sorry with the one(s) below it), */
	/* it enqueues the requests for those levels according to the usual */
	/* priority rules like any emulated interrupt */

	/* when the PDP-11 eventually lowers its priority enough to serve */
	/* the outstanding requests, E11's software grant service includes */
	/* issuing ioctl() calls to probe the outstanding BRs -- only then */
	/* do we actually perform a bus grant on that level, because we're */
	/* now definitely ready to receive the vector */

	/* if new BRs come in at higher levels before the existing one(s) */
	/* are serviced, all that means is that we raise the board's prio */
	/* to match and add to the list of BRs that may need service */

	/* the flags that say "this level may need service" are kept in */
	/* separate variables (BR4-BR7) rather than a single bit mask, */
	/* because the C language has no standard way to specify that */
	/* atomic operations be used to set and clear bits in those masks, */
	/* which could lead to race conditions on SMP machines when the ISR */
	/* is running on one CPU and the ioctl() handler on another */

	/* putting them in separate variables means they are set and cleared */
	/* with simple loads and stores, which are of course atomic */

	/* find highest active BR level */
	n=ioread32(BA->regs+BCIBMR);	/* read maint reg */
	BCIDELAY;
	if(n&0x400) {			/* BR7 is highest */
		n=7<<5;			/* lock out all ints */
		goto setbr7;
	}
	else if(n&0x200) {		/* BR6 */
		n=(6<<5)|1;		/* lock out BR4-BR6 */
		goto setbr6;
	}
	else if(n&0x100) {		/* BR5 */
		n=(5<<5)|1;		/* lock out BR4-BR5 */
		goto setbr5;
	}
	else if(n&0x80) {		/* BR4 */
		n=(4<<5)|1;		/* lock out BR4 */
		goto setbr4;
	}
	else return(IRQ_HANDLED);	/* must have changed their minds */

setbr7:	BA->BR7=0x80;			/* set our bit + all lower ones */
setbr6:	BA->BR6=0x40;
setbr5:	BA->BR5=0x20;
setbr4:	BA->BR4=0x10;

	iowrite32(n,BA->regs+BCIBPR);	/* disable this level and lower */
	BCIDELAY;

	/* interrupt owner if enabled */
	/* N.B. SIGIO value is ignored, signal is always SIGIO */
	if(BA->fasync_queue)
		kill_fasync(&BA->fasync_queue,SIGIO,POLL_PRI);

	return(IRQ_HANDLED);		/* it was ours */
}

/* read() call, currently unused */
static ssize_t bciread(struct file *filp,char *p,size_t n,loff_t *o)
{
	return(-EINVAL);
}

/* write() call, currently unused */
static ssize_t bciwrite(struct file *filp,const char *p,size_t n,loff_t *o)
{
	return(-EINVAL);
}

/* refresh DMA map hardware to match offset and pages array */
static int bcirefmap(struct barec *ba)
{
	unsigned int i, j, ptr, first;

	if(ba->soft)		/* no op if no hardware */
		return(0);

	if(!ba->mapped)		/* works only on BCI-2x04 */
		return(-EINVAL);

// we're not supposed to use virt_to_bus any more (asm/io.h just defines it as
// virt_to_phys anyway, on x86), but it's not at all clear what to use
// instead if your region is >128 KB and can't be allocated with
// dma_alloc_coherent -- well I could just make a jillion separate calls to
// dma_alloc_coherent
	first=virt_to_bus((void *)ba->pages[ba->offset>>PAGE_SHIFT]);
				/* get b31 of first page */

	/* count up 512 bytes at a time, since BCI uses VAX page size */
	for(i=0,ptr=ba->offset;i<(8192*8);i+=8,ptr+=512) {
		/* set BCI page "i" to relative address "ptr" */
		if(ptr<DMASIZE&&ba->pages[ptr>>PAGE_SHIFT]) {
			/* page exists, get its base address */
			j=virt_to_bus((void *)ba->pages[ptr>>PAGE_SHIFT]);

			/* b31 must be the same for all pages */
			/* (hardware limitation, bad on systems with >2 GB) */
			if(((j^first)&0x80000000)!=0) return(-EFAULT);

			/* compose bits 31:9 of address, add PVAL bit */
			j=((j				/* base of page */
				|(ptr&~PAGE_MASK))	/* + offset in page */
					>>9)		/* b31:9 => b22:0 */
				|0x80000000;		/* PVAL=1 */
		}
		else j=0x40000000;	/* page not present, set PNP */

		/* write scatter/gather map registers */
		/* first:  b31=PVAL, b30=PNP, b22:0=A31:9 */
		iowrite32(j,ba->regs+BCISGM+i);
		BCIDELAY;

		/* second word is unused but could be 32 more address bits */
		/* in future hardware with 64-bit addressing */
		iowrite32(0,ba->regs+BCISGM+i+4); /* MSW = 0 */
		BCIDELAY;
	}

	/* set b31 of DMA area from first page */
	outl((first&0x80000000)|0x00000001,ba->port+PLXDMC);
					/* top of RMAP=b31, DMAE=1 */
	BCIDELAY;

	/* enable or disable Q/Unibus DMA depending on map size */
	/* if we've mapped even one page, turn on DMA */
	if(ba->pages[ba->offset>>PAGE_SHIFT]!=0)
		ba->csr|=0x40;		/* pg 0 valid, set DEN */
	else ba->csr&=~0x40;		/* no memory, clear DEN */
	iowrite32(ba->csr,ba->regs+BCICSR);
	BCIDELAY;

	return(0);	/* happy */
}

/* enable DMA into user memory (which we'll lock) */
static int bciuserdma(struct barec *ba,void *addr,unsigned int len)
{
	int i, n, rc;

	if(ba->soft) goto happy;  /* no DMA on fake bus adapter */

	n=(len+PAGE_SIZE-1)/PAGE_SIZE;	/* get # (partial) pages */

	/* tell kernel to lock down memory for direct I/O */
	down_read(&current->mm->mmap_sem);
	rc=get_user_pages(current,current->mm,(unsigned long)addr,n,1,0,
		ba->pgs,	/* struct page buffer */
		NULL);		/* struct vma_area_struct buf (N/A) */
	up_read(&current->mm->mmap_sem);
	if(rc<0) goto fail;	/* failed */

	if(rc<n) {		/* didn't get all the pages we asked for */
		ba->npgs=rc;	/* unlock what we got */
		rc=(-ENOMEM);	/* complain about the rest */
		goto unlk;
	}
	ba->npgs=n;

	/* build scatter/gather descriptor list, program IOMMU if any */
	for(i=0;i<n;i++) {
		ba->sg[i].page=ba->pgs[i];  /* use next page */
		ba->sg[i].offset=0;	/* start at begn of page */
		ba->sg[i].length=PAGE_SIZE;  /* use entire page */
	}
	if(len&(PAGE_SIZE-1))	/* trim final page to actual length */
		ba->sg[i-1].length=len&(PAGE_SIZE-1);

	/* map scatter/gather operation */
	/* this just fetches pointers on x86, might allocate and set up */
	/* IOMMU registers on other architectures -- which is great, and */
	/* would even make the BCI-2x03s work if the DMA area appears as */
	/* contiguous DMA space, but if this system uses bounce buffers then */
	/* none of this will work since E11 memory won't see any of it, plus */
	/* E11 memory will get trashed when they DETACH BCIx: */
	rc=dma_map_sg(&ba->pcidev->dev,ba->sg,n,DMA_BIDIRECTIONAL);

// see if it succeeded (can it fail or do we just wait until regs free up?)
// then program the hardware with the correct map register values

/// page 451 of the LDD3 book describes this stuff
/// each of RC slots contains:
///		sg_dma_address(&(ba->sg[i])) => DMA addr of segment
///		sg_dma_len(&(ba->sg[i])) => length of segment

happy:	rc=0;			/* success */
fail:	return(rc);

unlk:	for(i=0;i<ba->npgs;i++)	/* unlock whatever user pages we got */
		page_cache_release(ba->pgs[i]);  /* (known not dirty) */
	ba->npgs=0;
	return(rc);
}

/* stop DMA into user memory, and unlock user memory */
static void bcinouserdma(struct barec *ba)
{
	unsigned int i;

	if(ba->npgs==0) return;		/* nothing to do */

	/* undo DMA mapping */
	/* no op on x86, may free UMRs (or whatever they're called here) on */
	/* other architectures */
	dma_unmap_sg(&ba->pcidev->dev,ba->sg,ba->npgs,DMA_BIDIRECTIONAL);

	/* unlock our pages */
	for(i=0;i<ba->npgs;i++) {
		/* we don't know if it's dirty, so say it is */
		if(!PageReserved(ba->pgs[i]))
			SetPageDirty(ba->pgs[i]);
		/* unlock it */
		page_cache_release(ba->pgs[i]);
	}
	ba->npgs=0;			/* no pages locked */
}

/* ioctl() call */
/* this is where we do our actual I/O */
static int bciioctl(struct inode *inode,struct file *filp,unsigned int cmd,
	unsigned long parm)
{
	struct barec *ba=filp->private_data;
	unsigned int n, t;
	int rc;
	void *p;

	/* dispatch based on command opcode */
	switch(cmd) {

	/* send reset pulse */
	case BCIRST:
		if(ba->soft) return(0);
		iowrite32(ba->csr|0x204,ba->regs+BCICSR);  /* set RST bit */
						/* (and write GER to clear) */
		udelay(100);			/* wait 100 usec (blocking) */
						/* (so no BCIDELAY needed) */
		ba->BR4=ba->BR5=ba->BR6=ba->BR7=0;  /* clear all BRs */
		iowrite32(ba->csr|0x200,ba->regs+BCICSR);  /* clear RST bit */
		BCIDELAY;
		return(0);

	/* set offset of DMA buffer from start of mmap() region */
	case BCIOFFSET:
		rc=-EINVAL;
		if((parm&(PAGE_SIZE-1))!=0)	/* must be page multiple */
			goto fail;
		if(parm>=DMASIZE) goto fail;	/* must fit within our mem */
		ba->offset=parm;		/* save */
		rc=bcirefmap(ba);		/* refresh map */
		if(rc<0) goto fail;
		return(ba->qbus);		/* return 1 if Q-bus */

	/* get bitmap of requesting BRs */
	case BCIREQS:
		return(ba->BR7|ba->BR6|ba->BR5|ba->BR4);

	/* attempt to grant interrupt */
	/* parm = CPU priority (4-7) */
	case BCIGRANT:
		if(ba->soft) goto timout;	/* no hardware, timeout */
		rc=-EINVAL;
		if((parm<4)||(parm>7)) goto fail;
		ba->br4[parm-4]=0;		/* remember we did it */
		parm=(parm-1)<<5;		/* CPL for one level lower */
		iowrite32(parm,ba->regs+BCIBPR);  /* set priority */
		BCIDELAY;
		iowrite32(parm|2,ba->regs+BCIBPR);  /* grant it */
		BCIDELAY;

		/* poll every 2 usec until GBSY clears */
		t=50;				/* give up after 50x */
		do {
			udelay(2);
			n=ioread32(ba->regs+BCIBVR);
		} while((--t)&&(n&0x0100));
		BCIDELAY;

		iowrite32(parm|1,ba->regs+BCIBPR);  /* enable this BR and up */
		BCIDELAY;
		if((n&0x80000100)!=0x80000000) {
						/* passive release if GBSY */
			iowrite32(ba->csr|0x200,ba->regs+BCICSR);
						/* write GER to clear it */
			goto timout;		/* still set or VAV clear */
		}

		return((n&0xFF)<<2);		/* return vector */

	/* data in */
	case BCIDATI:
		if(ba->soft) goto timout;	/* no hardware, timeout */
		/* write address reg, with C0=C1=0 for DATI */
		iowrite32(parm&017777777,ba->regs+BCIBAR);
		/* (no BCIDELAY needed because of udelay() in loop) */

		/* poll every 2 usec until DONE comes on */
	dati:	do {
			udelay(2);
			n=ioread32(ba->regs+BCIBDR);
		} while(!(n&0x80000000));
		BCIDELAY;

		if(n&0x00010000) goto timout;  /* bus timeout */

		return(n&0xFFFF);  /* otherwise happy */

	/* data in and pause (will DATO(B) immediately) */
	case BCIDATIP:
		if(ba->soft) goto timout;	/* no hardware, timeout */
		/* write addr reg, C0=1/C1=0 => DATIP */
		iowrite32((parm&017777777)|020000000,ba->regs+BCIBAR);
		goto dati;	/* otherwise same as DATI */
		/* (incl. no BCIDELAY needed because of udelay() in loop) */

	/* data out (word) */
	case BCIDATO:
		if(ba->soft) goto timout;	/* no hardware, timeout */
		/* write data reg */
		rc=-EFAULT;
		if(get_user(n,&((struct iopreq *)parm)->data)) goto fail;
		iowrite32(n,ba->regs+BCIBDR);
		BCIDELAY;

		/* write addr reg, C0=0/C1=1 => DATO */
		if(get_user(n,&((struct iopreq *)parm)->addr)) goto fail;
		iowrite32((n&017777777)|040000000,ba->regs+BCIBAR);
		/* (no BCIDELAY needed because of udelay() in loop) */

		/* poll every 2 usec until DONE comes on */
		do {
			udelay(2);
			n=ioread32(ba->regs+BCIBDR);
		} while(!(n&0x80000000));
		BCIDELAY;

		if(n&0x00010000) goto timout;
					/* bus timeout if XER set too */

		return(0);	/* happy */

	/* data out (byte) */
	case BCIDATOB:
		if(ba->soft) goto timout;	/* no hardware, timeout */
		/* write data reg */
		rc=-EFAULT;
		if(get_user(n,&((struct iopreq *)parm)->data)) goto fail;
		iowrite32(n,ba->regs+BCIBDR);
		BCIDELAY;

		/* write addr reg, C0=C1=1 => DATOB */
		if(get_user(n,&((struct iopreq *)parm)->addr)) goto fail;
		iowrite32((n&017777777)|060000000,ba->regs+BCIBAR);
		/* (no BCIDELAY needed because of udelay() in loop) */

		/* poll every 2 usec until DONE comes on */
		do {
			udelay(2);
			n=ioread32(ba->regs+BCIBDR);
		} while(!(n&0x80000000));
		BCIDELAY;

		if(n&0x00010000) goto timout;
					/* bus timeout if XER set too */

		return(0);	/* happy */

	/* enable/disable Unibus map */
	case BCIUMAP:
		if(!ba->soft) {
			iowrite32((parm?0x20:0),ba->regs+BCIMCR);
			BCIDELAY;
		}

		return(0);	/* happy */

	/* write a Unibus map register */
	case BCIWUMR:
		if(ba->soft) goto timout;	/* no hardware, timeout */
		rc=-EFAULT;
		if(get_user(t,&((struct umrreq *)parm)->reg)) goto fail;
		if(get_user(n,&((struct umrreq *)parm)->data)) goto fail;

		iowrite32(n&0177776L,ba->regs+BCIUMR+(t*4));  /* low word */
		BCIDELAY;		/* value doesn't stick w/o this! */
		iowrite32((n>>16)&077L,ba->regs+BCIUMR+(t*4)+2);  /* high */
		BCIDELAY;

		return(0);	/* happy */

	/* set actual priorities of BRA/BRB lines in Unibus adapters */
	/* (not enough conductors for all 4 sets of BRs/BGs so we get any 2) */
	case BCIBRAB:
		/* fetch parameters */
		rc=-EFAULT;
		if(get_user(t,&((struct brabrb *)parm)->bra)) goto fail;
		if(get_user(n,&((struct brabrb *)parm)->brb)) goto fail;

		/* check ranges */
		rc=-EINVAL;
		if(((t<4)||(t>6))||((n<5)||(n>7))) goto fail;

		/* write to hardware, if Unibus (Q-bus BCIs have all 4) */
		if((!ba->soft)&&(!ba->qbus)) {
			/* real, Unibus hardware */
			ba->csr=(ba->csr&~0x3018)	/* clear old BRA/BRB */
				|((t-4)<<3)		/* insert BRA */
				|((n-5)<<12);		/* and BRB */
			iowrite32(ba->csr,ba->regs+BCICSR);
			BCIDELAY;
		}
		return(0);	/* happy */

	/* EXPERIMENTAL: (not currently used by E11) */
	/* arrange DMA into user memory in place (instead of using our buf) */
	/* this is what we really wanted all along anyway, since it means we */
	/* can have >1 DMA device attached to E11 at a time */
	case BCIDMA:
		/* get address, length from user data structure */
		rc=-EFAULT;
		if(get_user(p,&((struct dmareg *)parm)->addr)) goto fail;
		if(get_user(n,&((struct dmareg *)parm)->len)) goto fail;

		/* make sure it starts on a page boundary */
		rc=-EINVAL;
		if(((unsigned long)p&(PAGE_SIZE-1))!=0)
			goto fail;		/* base must be page mult */
		rc=bciuserdma(ba,p,n);
		return(rc);

	default:
		rc=-EINVAL;
		goto fail;
	}

timout:	rc=-ETIME;			/* error = bus timeout */
fail:	return(rc);			/* return error */
}

/* routine called on user page fault within our DMA area */
static struct page *bcinopage(struct vm_area_struct *vma,unsigned long address,
	int *type)
{
	struct barec *ba;
	unsigned long addr;
	struct page *page;

	ba=vma->vm_file->private_data;	/* pick up our BA rec pointer */

	/* figure out page # within DMA buffer */
	/* (vm_pgoff is in pages) */
	addr=ba->pages[((address-vma->vm_start)>>PAGE_SHIFT)+vma->vm_pgoff];
	if(addr==0) goto fail;	/* missing page */
	page=virt_to_page(addr);  /* exists, get structure */
	get_page(page);		/* take the page */
	if(type!=NULL) *type=VM_FAULT_MINOR;  /* ("minor" = didn't need to be */
				/* swapped in from disk) */
	return(page);		/* return structure */

fail:	return(NOPAGE_OOM);	/* missing page */
}

/* update module use count based on mmap() operations too */
/* (not any more, use count is managed by kernel) */
static void bcivmaopen(struct vm_area_struct *vma)
{
	/* nothing to do any more */
}

/* close down mmap() */
static void bcivmaclose(struct vm_area_struct *vma)
{
	unsigned int i;
	struct barec *ba;

	/* flush any memory we have allocated */
	ba=vma->vm_file->private_data;
	for(i=0;i<MAXPAGE;i++) {
		if(ba->pages[i]!=0) {
			free_page(ba->pages[i]);
			ba->pages[i]=0;
		}
	}

	/* disable DMA */
	if(!ba->soft) {
		/* disable DMA */
		ba->csr&=~0x40;		/* clear DEN */
		iowrite32(ba->csr,ba->regs+BCICSR);
		BCIDELAY;			/* clear DEN, write it back */
	}
}

static struct vm_operations_struct bcivmops = {
	.open=	bcivmaopen,
	.close=	bcivmaclose,
	.nopage=bcinopage,
	/* other slots all NULL */
};

/* mmap() call */
/* allocate a DMA buffer of the requested size */
static int bcimmap(struct file *filp,struct vm_area_struct *vma)
{
	int rc;
	unsigned int i;
	size_t len;
	struct barec *ba;
	unsigned long ptr;

	len=vma->vm_end-vma->vm_start;

	rc=-ENOMEM;
	if(len>DMASIZE) goto fail;	/* make sure not too long */

	ba=filp->private_data;		/* get pointer to BA rec */

	/* modify DMA buffer size to whatever they asked for */
	/* (go through each possible page, add missing ones, delete extras) */
	for(ptr=0,i=0;ptr<DMASIZE;ptr+=PAGE_SIZE,i++) {
		if((ptr+PAGE_SIZE)<=len) {	/* should be part of memory */
			if(ba->pages[i]==0)	/* isn't yet */
				if((ba->pages[i]=__get_free_page(GFP_KERNEL))
					==0) goto fail;
		}
		else {		/* should not be memory */
			if(ba->pages[i]!=0) {	/* but it is */
				free_page(ba->pages[i]);  /* so free it */
				ba->pages[i]=0;
			}
		}
	}

	rc=bcirefmap(ba);	/* remap DMA hardware */

	vma->vm_ops=&bcivmops;	/* override */
	vma->vm_flags|=VM_RESERVED;  /* don't swap out */
/// this stuff is from the Rubini book:
//	vma->vm_inode=inode;
//// no longer exists, what do we do???
//	inode->i_count++;
/// same deal, no longer passed as an argument
/// seems to work fine w/o either of these ???
	bcivmaopen(vma);

fail:	return(rc);
}

/* open() call */
static int bciopen(struct inode *inode,struct file *filp)
{
	struct barec *ba;
	register unsigned int n;
	int rc;

	n=iminor(inode);		/* unit # = minor device # */

	/* look up bus adapter rec based on unit # */
	for(ba=balist;ba!=NULL&&n--;ba=ba->next) ;
	rc=-ENODEV;
	if(ba==NULL) goto fail;		/* walked off end of list */

	/* found it, make sure it's available */
	rc=-EBUSY;
	if(ba->inuse) goto fail;

	/* init hardware */
	ba->fasync_queue=NULL;	/* no async int handlers yet */
	if(!ba->soft) {
		/* no interrupts yet, so no BRs need probing */
		ba->BR4=ba->BR5=ba->BR6=ba->BR7=0;

		/* init */
		outl(0x900,ba->port+PLXICSR);
				/* set PLIE, PIE to enable interrupts */
		BCIDELAY;
		/* leave UBA default priorities (BRA=4 BRB=5) until ioctl() */
		iowrite32(0x205,ba->regs+BCICSR);
				/* GER, RST, BEN, but no DEN until user asks */
		udelay(100);		/* wait 100 usec */
		ba->csr=0x01;		/* (remember starting value) */
		iowrite32(0x201,ba->regs+BCICSR);  /* clear RST */
		BCIDELAY;
		iowrite32(0,ba->regs+BCIMCR);  /* disable Unibus map */
		BCIDELAY;
		iowrite32(1,ba->regs+BCIBPR);  /* enable bus requests */
		BCIDELAY;
	}
	ba->offset=0;		/* DMA starts at begn of mmap() area */
				/* (E11 V4.0+ no longer uses prefix area) */

	/* success, formally allocate the bus adapter */
	ba->inuse=1;		/* mark it "in use" */
	filp->private_data=ba;	/* save BA rec for other entry points */

	rc=0;			/* happy */
fail:	return(rc);		/* sad */
}

/* close() call */
static int bcirelease(struct inode *inode,struct file *filp)
{
	struct barec *ba = filp->private_data;

	/* clear FASYNC notification queue, if any */
	while(ba->fasync_queue)
		(void)fasync_helper(-1,filp,0,&ba->fasync_queue);

	if(!ba->soft) {
		/* disable interrupts and DMA */
		outl(0,ba->port+PLXICSR);  /* no ints (mask off in PLX chip) */
		BCIDELAY;
		iowrite32(0,ba->regs+BCIBPR);
		BCIDELAY;		/* disable BRs at BCI side too */
		ba->csr=0;
		iowrite32(0x200,ba->regs+BCICSR);  /* disable bus, no DMA */
		BCIDELAY;
	}

	bcinouserdma(ba);	/* stop DMA to user memory */

	/* mark as unused, ready to be opened again */
	ba->inuse=0;
	filp->private_data=NULL;

	return(0);
}

/* fasync() call */
/* this all seems like a bit much since we only allow one open per device */
/* anyway, but there doesn't seem to be any other way to send a signal to */
/* our owner */
static int bcifasync(int fd,struct file *filp,int mode)
{
	struct barec *ba = filp->private_data;

	return(fasync_helper(fd,filp,mode,&ba->fasync_queue));
}

/* vectors for driver operations, called by kernel */
static struct file_operations bciops = {
	.read=	bciread,
	.write=	bciwrite,
	.ioctl=	bciioctl,
	.mmap=	bcimmap,
	.open=	bciopen,
	.release=  bcirelease,
	.fasync=  bcifasync,
	/* rest of structure is left all NULLs */
};

/* delete a BCI-2x0x record */
static void delbci(struct barec *ba)
{
	struct barec *b, *p;

	/* unlink from BALIST, if it got linked in */
	for(b=balist,p=NULL;b!=NULL;p=b,b=b->next) {
		if(b==ba) {
			if(p!=NULL) p->next=b->next;
			else balist=b->next;
			if(balist==NULL)  /* disable bus if none left */
				bus_unregister(&bcibus);
			break;
		}
	}

	if(ba->cdevreg)			/* unregister char driver */
		cdev_del(&ba->cdev);
	if(ba->regs!=NULL)		/* flush phys window */
		iounmap(ba->regs);
	if(ba->memreg)			/* flush mem-mapped space ownership */
		release_mem_region(ba->regphys,ba->reglen);
	if(ba->portreg)			/* flush I/O space ownership */
		release_region(ba->port,PLXREG);
	if(ba->irqreg)			/* unhook our IRQ's int vector */
		free_irq(ba->irq,ba);
	if(ba->classreg)		/* flush /sys/class stuff */
		CLASSDEVDESTROY(bciclass,MKDEV(BCIMAJOR,ba->unit));
	kfree(ba);			/* flush BA record */
}

/* create a new BCI-2x0x record and fill it in */
/* (dev = PCI dev pointer, or NULL for fake software BCI) */
static int addbci(struct pci_dev *dev)
{
	struct barec *ba;		/* new BALIST entry */
	struct barec *b, *p;		/* current, prev entries */
	int n;
	int rc;				/* return code */
	char *err;			/* error msg */

	/* alloc new bus adapter record */
	ba=kmalloc(sizeof(struct barec),GFP_KERNEL);
	rc=-ENOMEM;			/* assume alloc failed */
	err="allocating memory";
	if(ba==NULL) goto fail;		/* it did */
	memset(ba,0,sizeof(struct barec));  /* clear all flags and pointers */

	/* initialize it, get PCI parms */
	ba->pcidev=dev;			/* save struct device for this dev */

	if(dev==NULL)			/* soft bus adapter (no HW) */
		ba->soft=ba->qbus=ba->mapped=1;  /* fake a BCI-2104 */
	else {
		ba->devid=dev->device;  /* PCI device ID */
		ba->port=dev->resource[1].start&PCI_BASE_ADDRESS_IO_MASK;
					/* PLX base I/O port */
		ba->irq=dev->irq;	/* IRQ */

		/* hook our IRQ's vector */
		err="installing int service routine";
		rc=request_irq(ba->irq,bciisr,SA_SHIRQ,NAME,ba);
		if(rc<0) goto fail;
		ba->irqreg=1;

		/* map physical address window to Xilinx mem-mapped BCI regs */
		ba->regphys=dev->resource[2].start&PCI_BASE_ADDRESS_MEM_MASK;
		ba->reglen=(dev->device==BCI2x03)?SIZE2x03:SIZE2x04;

		/* first allocate the region */
		rc=-EBUSY;
		err="reserving memory-mapped registers";
		if(request_mem_region(ba->regphys,ba->reglen,NAME)==NULL)
			goto fail;
		ba->memreg=1;

		/* map the address window (no op if the kernel has a */
		/* permanent mapping to physical memory) */
		rc=-ENOMEM;
		err="mapping window to device regs";
		if((ba->regs=ioremap_nocache(ba->regphys,ba->reglen))==NULL)
			goto fail;

		/* grab our I/O ports */
		/* since 2.4.2, skip check_region() and just do */
		/* request_region() and see if it succeeded */
		rc=-EBUSY;
		err="reserving I/O-mapped registers";
		if(request_region(ba->port,PLXREG,NAME)==NULL) goto fail;
		ba->portreg=1;

		/* init CSR (clear all, write GER=1 to clear it too) */
		ba->csr=0;
		iowrite32(0x200,ba->regs+BCICSR);
		BCIDELAY;

		/* detect Q-bus or Unibus (PCI ID is the same both ways) */
		ba->qbus=((ioread32(ba->regs+BCICSR)&0x8000)!=0);
		BCIDELAY;
		ba->mapped=(dev->device==BCI2x04);  /* page-mapped DMA */

		/* link BA rec to PCI device context */
		/* and make sure device has really gotten its resources */
		pci_set_drvdata(dev,ba);  /* link BA rec to PCI dev rec */
		rc=pci_enable_device(dev);  /* say hello to my little friend */
		if(rc) {
			pci_set_drvdata(dev,NULL);
			err="enabling PCI device";
			goto fail;
		}
		pci_set_master(dev);	/* enable bus master DMA */
	}

#if 1
	/* register bus if this is the first Q/Unibus adapter */
	if(balist==NULL) {
		rc=bus_register(&bcibus);  /* register Q/Unibus */
		err="registering bus type";
		if(rc<0) goto fail;
	}

	/* link to first missing slot in (sorted) BALIST */
	for(b=balist,p=NULL,n=0;b!=NULL;p=b,b=b->next,n++) {
		if(b->unit!=n)	/* gap in sequence, insert us here */
			break;
	}
	ba->unit=n;		/* save first free unit # */
	if(p!=NULL) p->next=ba;	/* link to prev if exists */
	else balist=ba;		/* otherwise we're first */
	ba->next=b;		/* link to next if exists (or NULL if last) */
#else
	/* link to end of BALIST, register bus if first */
	n=0;			/* init unit # */
	if(balist==NULL) {	/* first device */
		rc=bus_register(&bcibus);  /* register Q/Unibus */
		err="registering bus type";
		if(rc<0) goto fail;
		balist=ba;		/* we're the whole list */
	}
	else {	/* find end of list, count up # of entries */
		for(b=balist;n++,b->next!=NULL;b=b->next) ;
		b->next=ba;		/* link us to last one */
	}
#endif

	/* register separately for each unit number */
	/* (we don't know how many units we have until the system enumerates */
	/* all our PCI cards, and anyway on most systems there's just 1) */
	cdev_init(&ba->cdev,&bciops);	/* init our struct cdev */
	ba->cdev.owner=THIS_MODULE;
//	ba->cdev.ops=(&bciops);
// LDD3 book has this in example, but cdev_init() does it, since 2.6.1 at least
	rc=cdev_add(&ba->cdev,MKDEV(BCIMAJOR,n),1);  /* register this unit */
	err="registering device driver";
	if(rc<0) goto fail;
	ba->cdevreg=1;

	/* register our class (so udev will notice us) */
	/* will appear under /sys/class/uba */
	CLASSDEVCREATE(bciclass,MKDEV(BCIMAJOR,n),"bci%d",n);
	ba->classreg=1;

	/* tell them what we found (read with dmesg) */
	printk(KERN_NOTICE " /dev/logicalco/bci/%d: BCI-2%c0%c%s\n",
		n,ba->qbus?'1':'0',ba->mapped?'4':'3',
		ba->soft?" (software)":"");

	return(0);		/* success */

fail:	/* error, release any resources we'd allocated */
	printk(KERN_ERR NAME ":  error %s\n",err);
	if(ba!=NULL) delbci(ba);
	return(rc);		/* (caller may not check RC, see dmesg) */
}

/* BCI newly noticed */
/* (called only during pci_module_init() unless you've found some magical */
/* way to hot-plug BCI boards) */
static int __devinit bciprobe(struct pci_dev *dev,
	const struct pci_device_id *id)
{
	return(addbci(dev));
}

/* BCI unplugged (normally never called, but you never know) */
static void __devexit bciremove(struct pci_dev *dev)
{
	struct barec *ba;

	ba=pci_get_drvdata(dev);
	pci_set_drvdata(dev,NULL);

	delbci(ba);		/* shut it down, unlink, free memory */
}

static struct pci_driver bcidriver = {
	.name = NAME,
	.id_table = bcitable,
	.probe = bciprobe,
	.remove = __devexit_p(bciremove),
	/* no suspend/resume, BCI can't do it and Q/Unibus devs can't either */
};

/* module initialization entry point */
static int bci_init(void)
{
	printk(KERN_NOTICE "Driver for TLC BCI-2xxx bus adapters, V"
		BCIVERSION "\n");
  printk(KERN_NOTICE "Copyright (C) 2000-2006 by Digby's Bitpile, Inc.\n");

	/* create /sys class for our devices */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,2)
	/* 2.6.1 and earlier -- don't bother (use mknod) */
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,12)
	/* 2.6.2 - 2.6.12 */
	bciclass=class_simple_create(THIS_MODULE,"uba");
	if(IS_ERR(bciclass)) return(PTR_ERR(bciclass));
#else
	/* 2.6.13 and up */
	bciclass=class_create(THIS_MODULE,"uba");
	if(IS_ERR(bciclass)) return(PTR_ERR(bciclass));
#endif

	if(soft) (void)addbci(NULL);	/* add software BCI */
	return(pci_module_init(&bcidriver));
}

/* module cleanup entry point */
static void bci_exit(void)
{
	pci_unregister_driver(&bcidriver);

	/* go through entire list and delete all recs */
	while(balist!=NULL)
		delbci(balist);		/* flush BA record, unlink */

	/* destroy our /sys class entry */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,2)
	/* 2.6.1 and earlier -- don't bother */
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,12)
	class_simple_destroy(bciclass);	/* 2.6.2 - 2.6.12 */
#else
	class_destroy(bciclass);	/* 2.6.13 and up */
#endif
}

module_init(bci_init);
module_exit(bci_exit);
